/*
 * Simulator of MCS51 (cmdutil.cc)
 *
 * Copyright (C) 1997,97 Drotos Daniel, Talker Bt.
 * 
 * To contact author send email to drdani@mazsola.iit.uni-miskolc.hu
 *
 */

/* This file is part of MCS51 simulator: s51.

S51 is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

S51 is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with S51; see the file COPYING.  If not, write to the Free
Software Foundation, 59 Temple Place - Suite 330, Boston, MA
02111-1307, USA. */
/*@1@*/

#include "ddconfig.h"

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include "i_string.h"

#include "stypes.h"
#include "globals.h"
#include "uccl.h"


/*
 * Printing out an integer in binary format
 *____________________________________________________________________________
 *
 */

void
print_bin(long data, int bits, FILE *f)
{
  long mask= 1;

  mask= mask << ((bits >= 1)?(bits-1):0);
  while (bits--)
    {
      fprintf(f, "%c", (data&mask)?'1':'0');
      mask>>= 1;
    }
}


/*
 * Searching for a name in the specified table
 *____________________________________________________________________________
 *
 */

struct name_entry *
get_name_entry(struct name_entry tabl[], char *name, struct t_uc *uc)
{
  int i= 0;
  char *p;
  
  for (p= name; *p; *p= toupper(*p), p++);
  while (tabl[i].name &&
	 (!(tabl[i].cpu_type & uc->type) ||
	 (strcmp(tabl[i].name, name) != 0)))
    i++;
  if (tabl[i].name != NULL)
    return(&tabl[i]);
  else
    return(NULL);
}


/*
 * Interpreting a bitname
 *____________________________________________________________________________
 *
 */

bool
interpret_bitname(char *name, class t_uc *uc,
		  uchar **cell, uchar *celladdr,
		  uchar *bitaddr, uchar *bitmask,
		  char **symname)
{
  char *dot, *p;
  char *sym, bitnumstr[2];
  struct name_entry *ne;
  int bitnum, i;
  
  if ((dot= strchr(name, '.')) != NULL)
    {
      *dot++= '\0';
      if ((ne= get_name_entry(sfr_tab, name, uc)) == NULL)
	{
	  *celladdr= strtol(name, &p, 0);
	  if (p && *p)
	    {
	      dot--;
	      *dot= '.';
	      return(FALSE);
	    }
	}
      else
	*celladdr= ne->addr;
      if ((*celladdr < 0x20) ||
	  ((*celladdr > 0x2f) && (*celladdr < 0x80)) ||
	  ((*celladdr > 0x7f) && (*celladdr & 0x07)))
	return(FALSE);
      bitnum= strtol(dot, &p, 0);
      if ((p && *p) ||
	  (bitnum < 0) ||
	  (bitnum > 7))
	return(FALSE);
      if (*celladdr > 0x7f)
	*bitaddr= *celladdr + bitnum;
      else
	*bitaddr= (*celladdr - 0x20)*8 + bitnum;
      dot--;
      *dot= '.';
    }
  else
    {
      if ((ne= get_name_entry(bit_tab, name, uc)) == NULL)
	{
	  *bitaddr= strtol(name, &p, 0);
	  if ((p && *p) ||
	      (*bitaddr > 0xff))
	    return(FALSE);
	}
      else
	*bitaddr= ne->addr;
      if (*bitaddr > 0x7f)
	*celladdr= *bitaddr & 0xf8;
      else
	*celladdr= (*bitaddr >> 3) + 0x20;
    }
  // *bitaddr, *celladdr now OK
  *cell= uc->get_bit(*bitaddr);
  *bitmask= BIT_MASK(*bitaddr);
  // making symbolic name
  if (!symname)
    return(TRUE);
  i= 0;
  while (bit_tab[i].name &&
	 (bit_tab[i].addr != *bitaddr))
    i++;
  if (bit_tab[i].name)
    {
      sym= strdup(bit_tab[i].name);
      *symname= sym;
      return(TRUE);
    }
  i= 0;
  while (sfr_tab[i].name &&
	 (sfr_tab[i].addr != *celladdr))
    i++;
  if (sfr_tab[i].name)
    sym= strdup(sfr_tab[i].name);
  else
    {
      sym= (char *)malloc(3);
      sprintf(sym, "%02x", *celladdr);
    }
  sym= (char *)realloc(sym, strlen(sym)+2);
  strcat(sym, ".");
  sprintf(bitnumstr, "%1d", *bitaddr & 0x07);
  strcat(sym, bitnumstr);
  *symname= sym;
  return(TRUE);
}


/*
 * Processing escape sequencies in a string
 *____________________________________________________________________________
 *
 */

char *
proc_escape(char *string, int *len)
{
  char  spec_chars[]= "fnrtvab";
  char  spec[]= "\f\n\r\t\v\a\b";
  char  *s, *str, *p;

  s  = string;
  str= (char *)malloc(strlen(string)+1);
  p  = str;
  while (*s)
    {
      char *spec_c;

      if (*s == '\\' &&
	  *(s+1))
	{
	  s++;
	  if (*s == '0')
	    {
	      if (!isdigit(*(s+1)))
		{
		  *p++= '\0';
		  s++;
		}
	      else
		{
		  char *octal, *chk, data;
		  int i, j;
		  i= strspn(s, "01234567");
		  octal= (char *)malloc(i+1);
		  j= 0;
		  while (*s &&
			 (j < i))
		    octal[j++]= *s++;
		  octal[j]= '\0';
		  data= strtol(octal, &chk, 8);
		  if (!chk || !(*chk))
		    *p++= data;
		}
	    }
	  else
	    if ((spec_c= strchr(spec_chars, *s)) != NULL)
	      {
		*p++= spec[spec_c-spec_chars];
		s++;
	      }
	    else
	      *p++= *s++;
	}
      else
	*p++= *s++;
    }
  *p= '\0';
  *len= p-str;
  return(str);
}


/* End of cmdutil.cc */
