/*
 * Simulator of MCS51 (jmp.cc)
 *
 * Copyright (C) 1997,97 Drotos Daniel, Talker Bt.
 * 
 * To contact author send email to drdani@mazsola.iit.uni-miskolc.hu
 *
 */

/* This file is part of MCS51 simulator: s51.

S51 is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

S51 is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with S51; see the file COPYING.  If not, write to the Free
Software Foundation, 59 Temple Place - Suite 330, Boston, MA
02111-1307, USA. */
/*@1@*/

/* Bugs fixed by Sandeep Dutta:
 *	relative<->absolute jump in "jmp @a+dptr"
 */

#include "ddconfig.h"

#include <stdio.h>
#include <stdlib.h>

#include "uccl.h"
#include "stypes.h"


/*
 * 0x[02468ace]1 2 24 AJMP addr
 *____________________________________________________________________________
 *
 */

int
t_uc::inst_ajmp_addr(uchar code)
{
  uchar h, l;

  h= (code >> 5) & 0x07;
  l= fetch();
  tick(1);
  PC= (PC & 0xf800) | (h*256 + l);
  return(resGO);
}


/*
 * 0x10 3 12 JBC bit,addr
 *____________________________________________________________________________
 *
 */

int
t_uc::inst_jbc_bit_addr(uchar code)
{
  uchar bitaddr, *addr, jaddr;

  bitaddr= fetch();
  jaddr  = fetch();
  addr   = get_bit(bitaddr, &event_at.ri, &event_at.rs);
  if (read(addr) & BIT_MASK(bitaddr))
    {
      (*addr)&= ~BIT_MASK(bitaddr);
      PC= (PC + (signed char)jaddr) & (EROM_SIZE - 1);
    }
  return(resGO);
}


/*
 * 0x02 3 24 LJMP addr
 *____________________________________________________________________________
 *
 */

int
t_uc::inst_ljmp(uchar code)
{
  PC= fetch()*256 + fetch();
  tick(1);
  return(resGO);
}


/*
 * 0x[13579bdf]1 2 24 ACALL addr
 *____________________________________________________________________________
 *
 */

int
t_uc::inst_acall_addr(uchar code)
{
  uchar h, l, *sp;
  int res;

  h= (code >> 5) & 0x07;
  l= fetch();
  SFR[SP]++;
  sp= get_indirect(SFR[SP], &res);
  if (res != resGO)
    res= resSTACK_OV;
  (*sp)= PC & 0xff;
  tick(1);

  SFR[SP]++;
  sp= get_indirect(SFR[SP], &res);
  if (res != resGO)
    res= resSTACK_OV;
  (*sp)= (PC >> 8) & 0xff;
  PC= (PC & 0xf800) | (h*256 + l);
  return(res);
}


/*
 * 0x12 3 24 LCALL addr
 *____________________________________________________________________________
 *
 */

int
t_uc::inst_lcall(uchar code, uint addr)
{
  uchar h, l, *sp;
  int res;

  if (!addr)
    {
      h= fetch();
      l= fetch();
    }
  SFR[SP]++;
  sp= get_indirect(SFR[SP], &res);
  if (res != resGO)
    res= resSTACK_OV;
  (*sp)= PC & 0xff;
  if (!addr)
    tick(1);

  SFR[SP]++;
  sp= get_indirect(SFR[SP], &res);
  if (res != resGO)
    res= resSTACK_OV;
  (*sp)= (PC >> 8) & 0xff;
  if (addr)
    PC= addr;
  else
    PC= h*256 + l;
  return(res);
}


/*
 * 0x20 3 24 JB bit,addr
 *____________________________________________________________________________
 *
 */

int
t_uc::inst_jb_bit_addr(uchar code)
{
  uchar *addr, bitaddr, jaddr;

  addr= get_bit(bitaddr= fetch(), &event_at.ri, &event_at.rs);
  tick(1);
  jaddr= fetch();
  if (read(addr) & BIT_MASK(bitaddr))
    PC= (PC + (signed char)jaddr) & (EROM_SIZE-1);
  return(resGO);
}


/*
 * 0x22 1 24 RET
 *____________________________________________________________________________
 *
 */

int
t_uc::inst_ret(uchar code)
{
  uchar h, l, *sp;
  int res;

  sp= get_indirect(SFR[SP], &res);
  if (res != resGO)
    res= resSTACK_OV;
  h= *sp;
  SFR[SP]--;
  tick(1);

  sp= get_indirect(SFR[SP], &res);
  if (res != resGO)
    res= resSTACK_OV;
  l= *sp;
  SFR[SP]--;
  PC= h*256 + l;
  return(res);
}


/*
 * 0x30 3 24 JNB bit,addr
 *____________________________________________________________________________
 *
 */

int
t_uc::inst_jnb_bit_addr(uchar code)
{
  uchar *addr, bitaddr, jaddr;

  addr= get_bit(bitaddr= fetch(), &event_at.ri, &event_at.rs);
  tick(CYCLE_CLOCK);
  jaddr= fetch();
  if (!(read(addr) & BIT_MASK(bitaddr)))
    PC= (PC + (signed char)jaddr) & (EROM_SIZE-1);
  return(resGO);
}


/*
 * 0x32 1 24 RETI
 *____________________________________________________________________________
 *
 */

int
t_uc::inst_reti(uchar code)
{
  uchar h, l, *sp;
  int res;

  sp= get_indirect(SFR[SP], &res);
  if (res != resGO)
    res= resSTACK_OV;
  h= *sp;
  SFR[SP]--;
  tick(1);

  sp= get_indirect(SFR[SP], &res);
  if (res != resGO)
    res= resSTACK_OV;
  l= *sp;
  SFR[SP]--;
  PC= h*256 + l;

  was_reti= TRUE;
  if (it_levels->next)
    {
      struct it_level *p;
      p= it_levels->next;
      free(it_levels);
      it_levels= p;
    }
  return(res);
}


/*
 * 0x40 2 24 JC addr
 *____________________________________________________________________________
 *
 */

int
t_uc::inst_jc_addr(uchar code)
{
  uchar jaddr;

  jaddr= fetch();
  tick(1);
  if (GET_C)
    PC= (PC + (signed char)jaddr) & (EROM_SIZE-1);
  event_at.rs= ACC;
  return(resGO);
}


/*
 * 0x50 2 24 JNC addr
 *____________________________________________________________________________
 *
 */

int
t_uc::inst_jnc_addr(uchar code)
{
  uchar jaddr;

  jaddr= fetch();
  tick(1);
  if (!GET_C)
    PC= (PC + (signed char)jaddr) & (EROM_SIZE-1);
  event_at.rs= ACC;
  return(resGO);
}


/*
 * 0x60 2 24 JZ addr
 *____________________________________________________________________________
 *
 */

int
t_uc::inst_jz_addr(uchar code)
{
  uchar jaddr;

  jaddr= fetch();
  tick(1);
  if (!SFR[ACC])
    PC= (PC + (signed char)jaddr) & (EROM_SIZE-1);
  return(resGO);
}


/*
 * 0x70 2 24 JNZ addr
 *____________________________________________________________________________
 *
 */

int
t_uc::inst_jnz_addr(uchar code)
{
  uchar jaddr;

  jaddr= fetch();
  tick(1);
  if (SFR[ACC])
    PC= (PC + (signed char)jaddr) & (EROM_SIZE-1);
  return(resGO);
}


/*
 * 0x73 1 24 JMP @A+DPTR
 *____________________________________________________________________________
 *
 */

int
t_uc::inst_jmp_$a_dptr(uchar code)
{
  PC= (SFR[DPH]*256 + SFR[DPL] + SFR[ACC]) & (EROM_SIZE - 1);
  return(resGO);
}


/*
 * 0x80 2 24 SJMP addr
 *____________________________________________________________________________
 *
 */

int
t_uc::inst_sjmp(uchar code)
{
  PC= (PC + (signed char)fetch()) & (EROM_SIZE -1);
  tick(1);
  return(resGO);
}


/*
 * 0xb4 3 24 CJNE A,#data,addr
 *____________________________________________________________________________
 *
 */

int
t_uc::inst_cjne_a_$data_addr(uchar code)
{
  uchar data, jaddr;

  data = fetch();
  jaddr= fetch();
  tick(1);
  SET_C(SFR[ACC] < data);
  if (SFR[event_at.rs= ACC] != data)
    PC= (PC + (signed char)jaddr) & (EROM_SIZE - 1);
  return(resGO);
}


/*
 * 0xb5 3 24 CJNE A,addr,addr
 *____________________________________________________________________________
 *
 */

int
t_uc::inst_cjne_a_addr_addr(uchar code)
{
  uchar data, *addr, jaddr;

  addr = get_direct(fetch(), &event_at.ri, &event_at.rs);
  jaddr= fetch();
  tick(1);
  data= read(addr);
  SET_C(SFR[ACC] < data);
  if (SFR[event_at.rs= ACC] != data)
    PC= (PC + (signed char)jaddr) & (EROM_SIZE - 1);
  return(resGO);
}


/*
 * 0xb6-0xb7 3 24 CJNE @Ri,#data,addr
 *____________________________________________________________________________
 *
 */

int
t_uc::inst_cjne_$ri_$data_addr(uchar code)
{
  uchar *addr, data, jaddr;
  int res;

  addr = get_indirect(event_at.ri= *(get_reg(code & 0x01)), &res);
  data = fetch();
  jaddr= fetch();
  tick(1);
  SET_C(*addr < data);
  if (*addr != data)
    PC= (PC + (signed char)jaddr) & (EROM_SIZE - 1);
  return(res);
}


/*
 * 0xb8-0xbf 3 24 CJNE Rn,#data,addr
 *____________________________________________________________________________
 *
 */

int
t_uc::inst_cjne_rn_$data_addr(uchar code)
{
  uchar *reg, data, jaddr;

  reg  = get_reg(code & 0x07, &event_at.ri);
  data = fetch();
  jaddr= fetch();
  tick(1);
  SET_C(*reg < data);
  if (*reg != data)
    PC= (PC + (signed char)jaddr) & (EROM_SIZE - 1);
  return(resGO);
}


/*
 * 0xd5 3 24 DJNZ addr,addr
 *____________________________________________________________________________
 *
 */

int
t_uc::inst_djnz_addr_addr(uchar code)
{
  uchar *addr, jaddr;
  
  addr = get_direct(fetch(), &event_at.wi, &event_at.ws);
  jaddr= fetch();
  if (--(*addr))
    PC= (PC + (signed char)jaddr) & (EROM_SIZE-1);
  return(resGO);
}


/*
 * 0xd8-0xdf 2 24 DJNZ Rn,addr
 *____________________________________________________________________________
 *
 */

int
t_uc::inst_djnz_rn_addr(uchar code)
{
  uchar *reg, jaddr;
  
  reg  = get_reg(code & 0x07, &event_at.wi);
  jaddr= fetch();
  if (--(*reg))
    PC= (PC + (signed char)jaddr) & (EROM_SIZE-1);
  return(resGO);
}


/* End of jmp.cc */
