/*
 * Simulator of MCS51 (pobj.cc)
 *
 * Copyright (C) 1997,97 Drotos Daniel, Talker Bt.
 * 
 * To contact author send email to drdani@mazsola.iit.uni-miskolc.hu
 *
 */

/* This file is part of MCS51 simulator: s51.

S51 is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

S51 is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with S51; see the file COPYING.  If not, write to the Free
Software Foundation, 59 Temple Place - Suite 330, Boston, MA
02111-1307, USA. */
/*@1@*/

#include "ddconfig.h"

#include <stdio.h>
#include <stdlib.h>

#include "pstr.h"
#include "pobjt.h"
#include "pobj.h"


/*									    *
  ==========================================================================*
								    TObject *
  ==========================================================================*
									    *
*/

/* 
 * Initializing the object
 *_________________________________________________________________________
 *
 */

TObject::TObject(void) {}


/* 
 * Destructing the object: calling hte virtual Done method
 *_________________________________________________________________________
 *
 */

TObject::~TObject(void) {}


int TObject::Init(void) {return(0);}


/*									    *
  ==========================================================================*
								TCollection *
  ==========================================================================*
									    *
*/

/* 
 * Initializing a collection
 *_________________________________________________________________________
 *
 */

TCollection::TCollection(TIndex ALimit, TIndex ADelta):
TObject()
{
  Count= 0;
  Items= NULL;
  Limit= 0;
  Delta= ADelta;
  SetLimit(ALimit);
}


/* 
 * Disposing object's variables
 *_________________________________________________________________________
 *
 */

TCollection::~TCollection(void)
{
  delete Items;
}


/*
 * Get indexed item from the collection
 *_________________________________________________________________________
 *
 */

void *
TCollection::At(TIndex Index)
{
  if (Index < 0 || Index >= Count) Error(1, 0);
  return(Items[Index]);
}


/*
 * Deleting the indexed item from the collection
 *_________________________________________________________________________
 *
 */

void
TCollection::AtDelete(TIndex Index)
{
  if (Index < 0 || Index >= Count)
    Error(1, 0);
  Count--;
  memmove(&Items[Index], &Items[Index+1], (Count-Index)*sizeof(void *));
}


/*
 * Deleting the indexed item from the collection and disposing it
 *_________________________________________________________________________
 *
 */

void
TCollection::AtFree(TIndex Index)
{
void *Item= At(Index);

  AtDelete(Index);
  FreeItem(Item);
}


/*
 * Inserting a new item to the exact position
 *_________________________________________________________________________
 *
 */

void
TCollection::AtInsert(TIndex Index, void *Item)
{
  if (Index < 0 )
    Error(1, 0);
  if (Count == Limit)
    SetLimit(Count + Delta);

  memmove(&Items[Index+1], &Items[Index], (Count-Index)*sizeof(void *));
  Count++;

  Items[Index]= Item;
}


/* 
 * Put a new item to the collection. This function replaces an existing
 * item with a new one but it does not delete or dispose the old item!
 *_________________________________________________________________________
 *
 */

void
TCollection::AtPut(TIndex Index, void *Item)
{
  if (Index >= Count) Error(1, 0);
  Items[Index]= Item;
}


/*
 * Deleting an item from the collection but not disposing it
 *_________________________________________________________________________
 *
 */

void
TCollection::Delete(void *Item)
{
  AtDelete(IndexOf(Item));
}


/* 
 * Deleting all the items from the collection but not disposing them
 *_________________________________________________________________________
 *
 */

void
TCollection::DeleteAll(void)
{
  Count= 0;
}


/* 
 * Action taken when an error occure
 *_________________________________________________________________________
 *
 */

void
TCollection::Error(TIndex Code, TIndex Info)
{
  fprintf(stderr, 
	  "Collection index error. Code= %d, Info= %d.\n",
	  Code, Info);
  exit(Code);
}


/* 
 * Iterator method. This function calls 'Test' using every items as Test's
 * argument until Test returns TRUE.
 *_________________________________________________________________________
 *
 */

void *
TCollection::FirstThat(TTestFunc Test, void *Arg)
{
  for (TIndex i= 0; i < Count; i++)
    {
      if (Test(Items[i], Arg)) return(Items[i]);
    }
  return(NULL);
}


/* 
 * Iterator method. This function calls 'Action' using every items as
 * Action's argument.
 *_________________________________________________________________________
 *
 */

void
TCollection::ForEach(TAppFunc Action, void *Arg)
{
  for(TIndex i= 0; i < Count; i++) Action(Items[i], Arg);
}


/* 
 * Disposing an item.
 *_________________________________________________________________________
 *
 */

void
TCollection::FreeItem(void *Item)
{
  delete Item;
}


/* 
 * Get the number of collected items.
 *_________________________________________________________________________
 *
 */

int
TCollection::GetCount(void)
{
  return(Count);
}


/* 
 * Returning the index of an item.
 *_________________________________________________________________________
 *
 */

TIndex
TCollection::IndexOf(void *Item)
{
  for (TIndex i= 0; i < Count; i++)
    if (Item == Items[i])
      return(i);
  Error(1, 0);
  return(0);    /* Needed by Sun! */
}


/* 
 * Inserting a new item to the collection.
 *_________________________________________________________________________
 *
 */

TIndex
TCollection::Insert(void *Item)
{
TIndex loc= Count;

  AtInsert(Count, Item);
  return(loc);
}


/* 
 * Iterator method. This function calls 'Test' using every items
 * (in reversed order) as Test's argument until Test returns TRUE.
 *_________________________________________________________________________
 *
 */

void *
TCollection::LastThat(TTestFunc Test, void *Arg)
{
  for(TIndex i= Count; i > 0; i--)
    if (Test(Items[i-1], Arg))
      return(Items[i-1]);
  return(NULL);
}


/*
 * ???
 *_________________________________________________________________________
 *
 */

void
TCollection::Pack(void)
{
  void **CurDst= Items;
  void **CurSrc= Items;
  void **Last  = Items + Count;

  while (CurSrc < Last)
    {
      if (*CurSrc != NULL)
	*CurDst++= *CurSrc;
      *CurSrc++;
    }
}


/*
 * Setting up the maximum number of items. This function may expand
 * the size of the collection.
 *_________________________________________________________________________
 *
 */

void
TCollection::SetLimit(TIndex ALimit)
{
  void **AItems;

  if (ALimit < Count )
    ALimit= Count;
  if (ALimit > MaxCollectionSize)
    ALimit= MaxCollectionSize;
  if (ALimit != Limit)
    {
      if (ALimit == 0)
	AItems= NULL;
      else
	{
	  AItems = new void *[ALimit];
	  //i= ALimit*(sizeof(void *));
	  //AItems= (void **)malloc(i);
	  if (Count)
	    memcpy(AItems, Items, Count*sizeof(void *));
	}
      delete Items;
      Items= AItems;
      Limit= ALimit;
    }
}


/*									    *
  ==========================================================================*
							  TSortedcollection *
  ==========================================================================*
									    *
*/

/* 
 * Initilizing the sorted collection
 *_________________________________________________________________________
 *
 */

TSortedCollection::TSortedCollection(TIndex ALimit, TIndex ADelta):
  TCollection(ALimit, ADelta)
{
  Duplicates= FALSE;
}


TSortedCollection::~TSortedCollection(void) {}


/* 
 * Get the address of the key field in an item.
 *_________________________________________________________________________
 *
 */

void *
TSortedCollection::KeyOf(void *Item)
{
  return(Item);
}


/* 
 * Get index of an item.
 *_________________________________________________________________________
 *
 */

TIndex
TSortedCollection::IndexOf(void *Item)
{
  TIndex	i;

  if (Search(KeyOf(Item), i) == 0)
    return(ccNotFound);
  else
    {
      if (Duplicates)
	{
	  while (i < Count && Item != Items[i])
	    i++;
	}
      if (i < Count) return(i);
      else	     return(ccNotFound);
    }
}


/*
 * Inserting a new item to the collection
 *_________________________________________________________________________
 *
 */

TIndex
TSortedCollection::Insert(void *Item)
{
  TIndex i;

  if (Search(KeyOf(Item), i) == 0 || Duplicates)   // order dependency!
    AtInsert(i, Item);				   // must do Search
						   // before calling
						   // AtInsert
  return(i);
}


/*
 * Searching an item using binary search.
 *_________________________________________________________________________
 *
 */

bool
TSortedCollection::Search(void *Key, TIndex &Index)
{
  TIndex l  = 0;
  TIndex h  = Count - 1;
  bool   res= FALSE;
  
  while (l <= h)
    {
      TIndex i= (l + h) >> 1;
      TIndex c= Compare(KeyOf(Items[i]), Key);
      if (c < 0) l= i + 1;
      else
	{
	  h= i - 1;
	  if (c == 0)
	    {
	      res= TRUE;
	      if (!Duplicates)
		l= i;
	    }
	}
    }
  Index= l;
  return(res);
}


/*									    *
  ==========================================================================*
							  TStringcollection *
  ==========================================================================*
									    *
*/

/* 
 * Initilizing the string collection
 *_________________________________________________________________________
 *
 */

TStringCollection::TStringCollection(TIndex ALimit, TIndex ADelta):
  TSortedCollection(ALimit, ADelta)
{
  Duplicates= TRUE;
}


TStringCollection::~TStringCollection(void) {}


/* 
 * Comapare two string from the collection
 *_________________________________________________________________________
 *
 */

int
TStringCollection::Compare(void *Key1, void *Key2)
{
  return(strcmp((char *)Key1, (char *)Key2));
}


/* 
 * Deallocate string item of the collection
 *_________________________________________________________________________
 *
 */

void
TStringCollection::FreeItem(void* Item)
{
  delete Item;
}


/*									    *
  ==========================================================================*
							   TUnsortedStrings *
  ==========================================================================*
									    *
*/

/*
 * Initilizing the unsorted string collection
 *_________________________________________________________________________
 *
 */

TUnsortedStrings::TUnsortedStrings(TIndex ALimit, TIndex ADelta):
  TStringCollection(ALimit, ADelta)
{}


TUnsortedStrings::~TUnsortedStrings(void) {}


/*
 * Comapare two string from the collection
 *_________________________________________________________________________
 *
 */

int
TUnsortedStrings::Compare(void *Key1, void *Key2)
{
  return(-1);
}


/* 
 * Searching an item using linear search.
 *_________________________________________________________________________
 *
 */

bool
TUnsortedStrings::Search(void *Key, TIndex& Index)
{
TIndex i    = 0;
bool   found= FALSE;
void   *Actual;

  if ((Count) && (Key != NULL))
    {
      while (!found && (i < Count))
	{
	  Actual= KeyOf(At(i));
	  found = (Actual != NULL) &&
		  (Compare(Key, Actual) == 0);
	  i++;
	}
    }
  if (found) Index= i-1;
  else	     Index= Count;

  return(found);
}

/* End of pobj.cc */
